
def ascii2greek(x):
    if x.lower() in {'alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta', 'eta', 'theta', 'iota', 'kappa', 'lamda', 'mu', 'nu', 'xi', 'omicron', 'pi', 'rho', 'sigma', 'tau', 'upsilon', 'phi', 'chi', 'psi', 'omega'}:
        if x[0].isupper():
            x = eval("'\\N{GREEK CAPITAL LETTER %s}'" % x)
        else:
            x = eval("'\\N{GREEK SMALL LETTER %s}'" % x)
    return x

if __name__ == '__main__':
    ...


import numpy as np
import random

def extend(arr, mask, maxlength, padding_type):
    if isinstance(arr, (tuple, np.ndarray)):
        arr = [*arr]

    padding = [mask] * (maxlength - len(arr))
    if padding_type == 'tailing':
        arr.extend(padding)
    elif padding_type == 'leading':
        arr = padding + arr
    else:
        assert padding_type == 'general'
        for mask in padding:
            arr.insert(random.randrange(0, len(arr)), mask)

    return arr


def extend_tailing(arr, mask, maxlength):
    if isinstance(arr, tuple):
        arr = [*arr]

    padding = [mask] * (maxlength - len(arr))

    arr.extend(padding)

    return arr


def numpify_tailing(arr, mask_value=0):
    maxWidth = max(len(x) for x in arr)
    # arr is a 2-dimension array
    for i in range(len(arr)):
        arr[i] = extend_tailing(arr[i], mask_value, maxWidth)
    return np.array(arr)


def numpify_leading(arr, mask_value=0):
    return numpify(arr, mask_value, padding='leading')


def numpify(arr, mask_value=0, padding='tailing', dtype=None):
    '''
    
    :param arr:
    :param mask_value:
    :param shuffle: randomly insert the padding mask into the sequence！ this is used for testing masking algorithms!
    '''

    try:
        maxWidth = max(len(x) for x in arr)
    except (TypeError, AttributeError) as _:
        return np.array(arr)

    try:
        maxHeight = max(max(len(word) for word in x) for x in arr)
        for i in range(len(arr)):
            for j in range(len(arr[i])):
                arr[i][j] = extend(arr[i][j], mask_value, maxHeight, padding)
            arr[i] = extend(arr[i], [mask_value] * maxHeight, maxWidth, padding)
    except (TypeError, AttributeError, ValueError) as _:

        # arr is a 2-dimension array
        for i in range(len(arr)):
            arr[i] = extend(arr[i], mask_value, maxWidth, padding)

    arr = np.array(arr)
    if dtype:
        arr = arr.astype(dtype)
    return arr




from collections import deque

# https://signal-to-noise.xyz/post/bk-tree/
class BKTree:
    def __init__(self, distance_func):
        self._tree = None
        self._distance_func = distance_func

    def add(self, node):
        if self._tree is None:
            self._tree = (node, {})
            return

        current, children = self._tree
        while True:
            dist = self._distance_func(node, current)
            target = children.get(dist)
            if target is None:
                children[dist] = (node, {})
                break
            current, children = target

    def search(self, node, radius):
        if self._tree is None:
            return []

        candidates = deque([self._tree])
        result = []
        while candidates:
            candidate, children = candidates.popleft()
            dist = self._distance_func(node, candidate)
            if dist <= radius:
                result.append((dist, candidate))

            low, high = dist - radius, dist + radius
            candidates.extend(c for d, c in children.items()
                              if low <= d <= high)
        return result
    



import sys

class __LINE__:

    def __repr__(self):
        try:
            raise Exception
        except:
            return str(sys.exc_info()[2].tb_frame.f_back.f_lineno)

 
__LINE__ = __LINE__()



from std import computed, equal_range, binary_search

class SymbolicSet:
    is_EmptySet = False
    
    is_Range = False
    
    is_Union = False
    
    def symmetric_difference(self, that):
        return (self | that) - (self & that)


class EmptySet(SymbolicSet):
    is_EmptySet = True
    
    @property
    def card(self):
        return 0
    
    def __eq__(self, that):
        return that is None or that.is_EmptySet
    
    def __sub__(self, that):
        return self
    
    def __add__(self, offset):
        return self
    
    def __and__(self, offset):
        return self
    
    def __or__(self, that):
        return that

    def symmetric_difference(self, that):
        return that

    @property
    def dict(self):
        return {}

    def __str__(self):
        return '{}'

    __repr__ = __str__

    def __bool__(self):
        return False
    
    __nonzero__ = __bool__
    
    #lhs in self
    def __contains__(self, lhs):
        if lhs.is_EmptySet:
            return True
        
    def __iter__(self):  # Get iterator object on iter
        return RangeIterator(Range(0, 0))
    
class Union(SymbolicSet):
    is_Union = True
    
    @staticmethod
    def new(*arguments):
        if not arguments:
            return EmptySet()
            
        if len(arguments) == 1:
            return arguments[0]
            
        return Union(*arguments)
        
    def __init__(self, *args, **kwargs):
        if not args:
            args = kwargs['args']
            args = [Range(**arg) for arg in args]

        #assert not any(arg.is_EmptySet or arg.is_Union for arg in args)
        self.args = args
    
    @computed
    def card(self):
        card = 0
        for arg in self.args:
            card += arg.card
        return card
    
    def __and__(self, that):
        if that.is_EmptySet:
            return that

        args = []
        if that.is_Range:
            for i in range(*equal_range(self.args, that)):
                arg = self.args[i]
                arg &= that
                assert arg.is_Range
                    
                args.append(arg)
        else:
            for arg in self.args:
                arg &= that
                if arg.is_EmptySet:
                    continue
                    
                if arg.is_Union:
                    args += arg.args
                else:
                    args.append(arg)
            
        return Union.new(*args)
    
    def __eq__(self, that):
        if that and that.is_Union and len(self.args) == len(that.args):
            for i in range(len(self.args)):
                if self.args[i] != that.args[i]:
                    return False
            return True
    
    def __sub__(self, that):
        args = []
        for arg in self.args:
            arg -= that
            if arg.is_EmptySet:
                continue
                
            if arg.is_Union:
                args += arg.args    
            else:
                args.append(arg)
        
        if not args:
            return EmptySet()
            
        if len(args) == 1:
            return args[0]
            
        return Union(*args)
    
    def __add__(self, offset):
        return Union(*map(lambda el : el + offset, self.args))

    def __or__(self, that):
        if that.is_EmptySet:
            return self
        
        if that.is_Range:
            for i in range(len(self.args)):
                arg = self.args[i]
                arg |= that
                if arg.is_Range:
                    args = [*self.args]
                    del args[i]
                    if len(args) == 1:
                        return args[0] | arg
                    
                    return Union(*args) | arg
            
            args = [*self.args]
            
            index = binary_search(args, that)
            
            args.insert(index, that)
            
            return Union(*args)

        for arg in self.args:
            that |= arg
        return that

    def union_without_merging(self, that):
        if that.is_EmptySet:
            return self
        
        if that.is_Range:
            for arg in self.args:
                that -= arg
            
            if that.is_EmptySet:
                return self
            
            if that.is_Range:
                that = [that]
            else:
                that = that.args
                
            args = [*self.args]
            
            for that in that:
                index = binary_search(args, that)
                args.insert(index, that)
            
            return Union(*args)
        
        if that.is_Union:
            for arg in that.args:
                assert arg.is_Range
                self = self.union_without_merging(arg)
                
            return self
        raise
    
    @property
    def dict(self):
        return dict(args=[*map(lambda arg: arg.dict, self.args)])

    def __str__(self):
        return ' | '.join((str(arg) for arg in self.args))

    __repr__ = __str__

    #lhs in self
    def __contains__(self, lhs):
        if isinstance(lhs, int) or lhs.is_Range:
            return any(lhs in arg for arg in self.args)
        
        if lhs.is_EmptySet:
            return True
        
        if lhs.is_Union:
            return all(arg in self for arg in lhs.args)


class Range(SymbolicSet):
    is_Range = True
    
    def __init__(self, start, stop):
        self.start = start
        self.stop = stop
    
    @property
    def args(self):
        return [self.start, self.stop]
    
    def __and__(self, that):
        if that.is_Union:
            return that & self
        
        [start, stop] = that.args
        
        if start >= self.stop or stop <= self.start:
            return EmptySet()
            
        return Range(max(self.start, start), min(self.stop, stop))     

    def __eq__(self, that):
        if that and that.is_Range:
            return self.start == that.start and self.stop == that.stop
    
    def __or__(self, that):
        if that.is_Range:
            [start, stop] = that.args
            if start > self.stop:
                return Union(self, that)
                
            if stop < self.start:
                return Union(that, self)
                
            return Range(min(self.start, start), max(self.stop, stop))     
        
        if that.is_EmptySet:
            return self
            
        return that | self
    
    def union_without_merging(self, that):
        if that.is_EmptySet:
            return self
            
        if that.is_Range:
            [start, stop] = that.args
            if start >= self.stop:
                return Union(self, that)
                
            if stop <= self.start:
                return Union(that, self)
            
            mid = self & that
            assert mid
            
            lhs = self - that
            rhs = that - self
            
            if not rhs:
                return mid.union_without_merging(lhs)
            
            if not lhs:
                return mid.union_without_merging(rhs)
            
            if lhs.start > rhs.start:
                lhs, rhs = rhs, lhs
            return Union(lhs, mid, rhs)
        
        return that.union_without_merging(self)
    
    def __sub__(self, that):
        if that.start >= self.stop:
            return self
        
        if that.start > self.start:
            if that.stop >= self.stop:
                return Range(self.start, that.start)
                
            return Union(Range(self.start, that.start), Range(that.stop, self.stop))
        else:
            if that.stop >= self.stop:
                return EmptySet()
                
            if that.stop > self.start:
                return Range(that.stop, self.stop)
            
            return Range(self.start, self.stop)
    
    @computed
    def card(self):
        return self.stop - self.start
    
    def __add__(self, offset):
        return Range(self.start + offset, self.stop + offset)
    
    @property
    def dict(self):
        return dict(start=self.start, stop=self.stop)
            
    def __str__(self):
        return f"[{self.start}; {self.stop})"

    __repr__ = __str__
    
    #lhs in self
    def __contains__(self, lhs):
        if isinstance(lhs, int):
            return lhs >= self.start and lhs < self.stop
        
        if lhs.is_EmptySet:
            return True
        
        if lhs.is_Range:
            return lhs.start >= self.start and lhs.stop <= self.stop
        
        if lhs.is_Union:
            return all(arg in self for arg in lhs.args)
    
    def batches(self, step):
        start, stop = self.start, self.stop
        interval = []
        for i in range(start, stop, step):
            interval.append(Range(i, min(i + step, stop)))
                
        return interval        

    def __lt__(self, that):
        return self.stop <= that.start
    
    def __gt__(self, that):
        return self.start >= that.stop 

    def __iter__(self):  # Get iterator object on iter
        return RangeIterator(self)


class RangeIterator:
    def __init__(self, range):
        self.start = range.start
        self.stop = range.stop
        self.current = range.start
    
    def __next__(self):
        if self.current >= self.stop:
            raise StopIteration
        
        current = self.current
        self.current += 1
        return current
    

class Range2D(SymbolicSet):
    is_Range2D = True
    
    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
    
    @property
    def x_stop(self):
        return self.x + self.width

    @property
    def y_stop(self):
        return self.y + self.height

    @property
    def args(self):
        return [self.x, self.y, self.x_stop, self.y_stop]
    
    def __and__(self, that):
        if that.is_Union:
            return that & self
        
        x, y, x_stop, y_stop = that.args
        
        if x >= self.x_stop or x_stop <= self.x or y >= self.y_stop or y_stop <= self.y:
            return EmptySet()

        x = max(self.x, x)
        x_stop = min(self.x_stop, x_stop)
        
        y = max(self.y, y)
        y_stop = min(self.y_stop, y_stop)
        
        width = x_stop - x
        height = y_stop - y
        return Range2D(x, y, width, height)     

    def __eq__(self, that):
        if that and that.is_Range2D:
            return self.x == that.x and self.width == that.width and self.y == that.y and self.height == that.height
    
    def __or__(self, that):
        if that.is_Range2D:
            x, y, x_stop, y_stop = that.args
            if x == self.x and x_stop == self.x_stop:
                if y == self.y_stop:
                    return Range2D(self.x, self.y, self.width, self.height + that.height)
                if self.y == y_stop:
                    return Range2D(x, y, self.width, self.height + that.height)

            elif y == self.y and y_stop == self.y_stop:
                if x == self.x_stop:
                    return Range2D(self.x, self.y, self.width + that.width, self.height)
                if self.x == x_stop:
                    return Range2D(x, y, self.width + that.width, self.height)
                
            return Union(self, that)

        if that.is_EmptySet:
            return self
            
        return that | self
    
    def __sub__(self, that):
        if that.is_Range2D:
            if that.x >= self.x_stop or that.y >= self.y_stop or that.x_stop <= self.x or that.y_stop <= self.y:
                return self
            
            # that.x < self.x_stop && that.x_stop > self.x
            # that.y < self.y_stop && that.y_stop > self.y
            if that.x <= self.x:
                if that.x_stop < self.x_stop:
                    #that.x <= self.x && that.x_stop < self.x_stop
                    if that.y <= self.y:
                        if that.y_stop < self.y_stop:
                            #that.y <= self.y && that.y_stop < self.y_stop
                            return Union(
                                Range2D(self.x, that.y_stop, self.width, self.y_stop - that.y), 
                                Range2D(that.x_stop, self.y, self.x_stop - that.x_stop, that.y_stop - self.y))
                        else:
                            #that.y <= self.y && that.y_stop >= self.y_stop
                            return Range2D(that.x_stop, self.y, self.x_stop - that.x_stop, self.height)
                    else :
                        if that.y_stop < self.y_stop:
                            #that.y > self.y && that.y_stop < self.y_stop
                            return Union(
                                Range2D(self.x, self.y, self.width, that.y - self.y), 
                                Range2D(self.x, that.y_stop, self.width, self.y_stop - that.y_stop),
                                Range2D(that.x_stop, that.y, self.x_stop - that.x_stop, that.height))                            
                        else:
                            #that.y > self.y && that.y_stop >= self.y_stop
                            return Union(
                                Range2D(self.x, self.y, self.width, that.y - self.y),
                                Range2D(that.x_stop, that.y, self.x_stop - that.x_stop, self.y_stop - that.y))                            
                else :
                    #that.x <= self.x && that.x_stop >= self.x_stop
                    if that.y <= self.y:
                        if that.y_stop < self.y_stop:
                            #that.y <= self.y && that.y_stop < self.y_stop
                            return Range2D(self.x, that.y_stop, self.width, that.y_stop - self.y);
                        else:
                            #that.y <= self.y && that.y_stop >= self.y_stop
                            return EmptySet()
                    else :
                        if that.y_stop < self.y_stop:
                            #that.y > self.y && that.y_stop < self.y_stop
                            return Union(
                                Range2D(self.x, self.y, self.width, that.y - self.y), 
                                Range2D(self.x, that.y_stop, self.width, self.y_stop - that.y_stop))
                        else :
                            #that.y > self.y && that.y_stop >= self.y_stop
                            return Range2D(self.x, self.y, self.width, that.y - self.y)
            else :
                if that.x_stop < self.x_stop:
                    #that.x > self.x && that.x_stop < self.x_stop
                    if that.y <= self.y:
                        if that.y_stop < self.y_stop:
                            #that.y <= self.y && that.y_stop < self.y_stop
                            return Union(
                                Range2D(self.x, self.y, that.x - self.x, self.height), 
                                Range2D(that.x, that.y_stop, self.x_stop - that.x, self.y_stop - that.y_stop),
                                Range2D(that.x_stop, self.y, self.x_stop - that.x_stop, that.y_stop - self.y))
                        else :
                            #that.y <= self.y && that.y_stop >= self.y_stop
                            return Union(
                                Range2D(self.x, self.y, that.x - self.x, self.height), 
                                Range2D(that.x_stop, self.y, self.x_stop - that.x_stop, that.height))
                    else :
                        if that.y_stop < self.y_stop:
                            #that.y > self.y && that.y_stop < self.y_stop
                            return Union(
                                Range2D(self.x, self.y, self.width, that.y - self.y), 
                                Range2D(self.x, that.y, that.x - self.x, self.y_stop - that.y),
                                Range2D(that.x, that.y_stop, self.x_stop - self.x, self.y_stop - that.y_stop),
                                Range2D(that.x_stop, that.y, self.x_stop - that.x_stop, that.height))
                        else :
                            #that.y > self.y && that.y_stop >= self.y_stop
                            return Union(
                                Range2D(self.x, self.y, self.width, that.y - self.y), 
                                Range2D(self.x, that.y, that.x - self.x, self.y_stop - that.y),
                                Range2D(that.x_stop, that.y, self.x_stop - that.x_stop, self.y_stop - that.y))
                else :
                    #that.x > self.x && that.x_stop >= self.x_stop
                    if that.y <= self.y:
                        if that.y_stop < self.y_stop:
                            #that.y <= self.y && that.y_stop < self.y_stop
                            return Union(
                                Range2D(self.x, self.y, that.x - self.x, self.height), 
                                Range2D(that.x, that.y_stop, self.x_stop - that.x, self.y_stop - that.y_stop));
                        else:
                            #that.y <= self.y && that.y_stop >= self.y_stop
                            return Range2D(self.x, self.y, that.x - self.x, self.height); 
                    else:
                        if that.y_stop < self.y_stop:
                            #that.y > self.y && that.y_stop < self.y_stop
                            return Union(
                                Range2D(self.x, self.y, self.width, that.y - self.y), 
                                Range2D(self.x, that.y, that.x - self.x, self.y_stop - that.y),
                                Range2D(that.x, that.y_stop, self.x_stop - self.x, self.y_stop - that.y_stop))
                        else:
                            #that.y > self.y && that.y_stop >= self.y_stop
                            return Union(
                                Range2D(self.x, self.y, self.width, that.y - self.y), 
                                Range2D(self.x, that.y, that.x - self.x, self.y_stop - that.y))
                            
        elif that.is_EmptySet:
            return self
        elif that.is_Union:
            for region in that.args:
                self -= region
            
            return self
    
    @computed
    def card(self):
        return self.width * self.height
    
    def __add__(self, offset_x, offset_y):
        return Range2D(self.x + offset_x, self.y + offset_y, self.width, self.height);
    
    @property
    def dict(self):
        return dict(start=self.start, stop=self.stop)
            
    def __str__(self):
        return f"[{self.x}, {self.y}, {self.width}, {self.height}]"

    __repr__ = __str__
    
    #lhs in self
    def __contains__(self, lhs):
        if isinstance(lhs, (list, tuple)):
            x, y = lhs
            return x >= self.x and y < self.x_stop and y >= self.y and y < self.y_stop
        
        if lhs.is_EmptySet:
            return True
        
        if lhs.is_Range2D:
            x, y, x_stop, y_stop = self.args
            return x >= self.x and x_stop <= self.x_stop and y >= self.y and y_stop <= self.y_stop
        
        if lhs.is_Union:
            return all(arg in self for arg in lhs.args)
       
            
if __name__ == '__main__':
    ...
    



# to comply with the standards of java and mysql regex engine
import regex as re
from std import computed, Object, clip, splice, binary_search, toggleCase
import html

from std.sets import Range, Union

def add(arr, val):
    return [x + val for x in arr]

class XMLNode:
    is_XMLNodeCaret = False
    is_XMLNodeText = False
    is_XMLNodeBinaryTag = False
    is_XMLNodeSingleTag = False
    is_XMLNodeArray = False
    is_XMLNodeUnbalancedTag = False
    
    def __init__(self, parent=None):
        self.parent = parent

    @property
    def root(self):
        while self.parent:
            self = self.parent
        return self 

    def has_recursive_error(self):
        try:
            str(self.root)
        except Exception as e:
            return True
        
    def replace(self, *_):
        ...

    @computed
    def physicalText(self):
        caret = self
        s = ''
        while caret is not None:
            s += str(caret)
            caret = caret.parent
#         s += '\n'
        return s
    
    def __str__(self):
        return self.physicalText

    __repr__ = __str__

    def append_left_tag(self, tag):
        caret = XMLNodeCaret()
        parent = self.parent
        node = XMLNodeBinaryTag(tag, caret, parent=parent)
        if isinstance(parent, XMLNodeArray):
            parent.args.append(node)
        else:
            node = XMLNodeArray([self, node], parent)
            if parent:
                parent.replace(self, node)
        return caret

    def append_text_node(self, text):
        parent = self.parent
        node = XMLNodeText(text, parent)
        if isinstance(parent, XMLNodeArray):
            assert not parent.args[-1].is_XMLNodeText
            parent.args.append(node)
            return node
        else:
            array = XMLNodeArray([self, node], parent)
            if parent:
                parent.replace(self, array)
            return array

    def append_single_tag(self, tag):
        parent = self.parent
        node = XMLNodeSingleTag(tag, parent)
        if isinstance(parent, XMLNodeArray):
            parent.args.append(node)
            return node
        else:
            array = XMLNodeArray([self, node], parent)
            if parent:
                parent.replace(self, array)
            return array

    def enumerate(self):
        assert False, type(self)
        
    def has(self, cls):
        return False
    
    def __repr__(self):
        return str(self)
    
    @property
    def zeros(self):
        return []
    
    @computed
    def style(self):
        return {}

    def modify_style(self, tag):
        if self.text:
            set = Range(0, len(self.text))
            _style = self.style
            if tag in _style:
                _style[tag] = _style[tag].union_without_merging(set)
            else:
                _style[tag] = set
    
    @property
    def length(self):
        return self.stop - self.start
    
    def sanctity_check(self):
        ...
        
    @computed
    def style_traits(self):
        style_traits = [''] * len(self.text)
        for i, s in enumerate(self.style_input):
            s = [*s]
            s.sort()
            s = '.'.join(s)
            style_traits[i] = s
        return style_traits
    
    @computed
    def style_input(self):
        style_input = [set() for _ in range(len(self.text))]
        for tag, s in self.style.items():
            if s.is_Range:
                for i in range(s.start, s.stop):
                    style_input[i].add(tag)
            elif s.is_Union:
                for s in s.args:
                    for i in range(s.start, s.stop):
                        style_input[i].add(tag)
        return style_input
    

class XMLNodeCaret(XMLNode):
    is_XMLNodeCaret = True
    
    def append_left_tag(self, tag):
        caret = XMLNodeCaret()
        node = XMLNodeBinaryTag(tag, caret, parent=self.parent)
        if self.parent:
            self.parent.replace(self, node)
        return caret

    def append_text_node(self, text):
        node = XMLNodeText(text, self.parent)
        if self.parent:
            self.parent.replace(self, node)
        return node

    def append_single_tag(self, tag):
        node = XMLNodeSingleTag(tag, self.parent)
        if self.parent:
            self.parent.replace(self, node)
        return node

    @property
    def text(self):
        return ""
    
    @computed
    def physicalText(self):
        return ""
    
    @property
    def logicalLength(self):
        return 0
    
    @property
    def texts(self):
        return []
    
    def enumerate(self):
        yield

    def has(self, cls):
        return isinstance(self, cls)
    
    @property
    def length(self):
        return 0

    
class XMLNodeText(XMLNode):
    is_XMLNodeText = True
    
    def __init__(self, text, parent=None):
        super().__init__(parent)
        self.arg = text

    @property
    def start(self):
        return self.arg.start
    
    @property
    def stop(self):
        return self.arg.stop

    @start.setter
    def start(self, start):
        self.arg.start = start

    @stop.setter
    def stop(self, stop):
        self.arg.stop = stop
    
    @computed
    def text(self):
        return self.arg.text
    
    @computed
    def physicalText(self):        
        return str(self.arg)
    
    @property
    def logicalLength(self):
        return self.arg.length

    @computed
    def texts(self):
        return [self.text]
    
    def logical2physical(self, pos):
        return pos

    def physical2logical(self, pos):
        return pos
    
    def is_physically_valid(self, pos):
        return 0 <= pos < self.length

    def is_sibling(self, start, end):
        return 0 <= start < end <= self.length 
        
    def getPhysicalIndices(self, start, stop):
        return start, stop
        
    def enumerate(self):
        yield (self.lexeme, False)

    def has(self, cls):
        return isinstance(self, cls)

    def append_text_node(self, text):
        assert self.stop == text.start, "this.stop == text.start"
        self.stop = text.stop
        return self
    

class XMLNodeBinaryTag(XMLNode):
    is_XMLNodeBinaryTag = True
    
    def __init__(self, tagBegin, arg, tagEnd=None, parent=None):
        super().__init__(parent)
        self.tagBegin = tagBegin
        self.arg = arg
        self.tagEnd = tagEnd
        self.arg.parent = self

    def set_tagEnd(self, tagEnd):
        self.tagEnd = tagEnd
        if 'physicalText' in self.__dict__:
            del self.__dict__['physicalText']
        
    @property
    def tag(self):
        return self.tagBegin.tag
    
    @property
    def start(self):
        return self.tagBegin.start
    
    @property
    def stop(self):
        if self.is_unbalanced:
            if self.arg.is_XMLNodeCaret:
                return self.tagBegin.stop
            return self.arg.stop
        return self.tagEnd.stop
    
    @computed
    def physicalText(self):
        s = str(self.tagBegin) + str(self.arg)
        if self.is_unbalanced:
            return s
        return s + str(self.tagEnd)

    @computed
    def text(self):
        return self.arg.text

    @property
    def logicalLength(self):
        return self.arg.logicalLength

    @computed
    def texts(self):
        return self.arg.texts
    
    @computed
    def zeros(self):
        return self.arg.zeros if self.text else [0]
        
    @computed
    def style(self):
        _style = {}
        
        if self.text:
            if self.arg.is_XMLNodeArray:
                args = self.arg.args
                if self.tag in ("msub", "munder"):
                    if len(args) == 2:
                        args[1].modify_style('sub')
                            
                elif self.tag in ("msup", "mover"):
                    if len(args) == 2:
                        args[1].modify_style('sup')
                            
                elif self.tag in ("msubsup", "munderover"):
                    if len(args) == 3:
                        args[1].modify_style('sub')
                        args[2].modify_style('sup')
            
            _style[self.tag] = Range(0, len(self.text))
            for tag, set in self.arg.style.items():
                if tag in _style:
                    _style[tag] = _style[tag].union_without_merging(set)
                else:
                    _style[tag] = set
            
        return _style

    @property
    def is_unbalanced(self):
        return not self.tagEnd or isinstance(self.tagEnd, XMLNodeUnbalancedTag) 
    
    def logical2physical(self, pos):
        return self.arg.logical2physical(pos) + self.tagBegin.length

    def physical2logical(self, pos):
        return clip(self.arg.physical2logical(pos - self.tagBegin.length), 0, len(self.text) - 1)
    
    def is_physically_valid(self, pos):
        if self.tagBegin.length <= pos < self.length - self.tagEnd.length:
            return self.arg.is_physically_valid(pos - self.tagBegin.length)
    
    def is_sibling(self, start, end):
        if self.tagBegin.length <= start < end <= self.length - self.tagEnd.length:
            start -= self.tagBegin.length
            end -= self.tagBegin.length
            return self.arg.is_sibling(start, end)

    def getPhysicalIndices(self, start, stop):
        start, stop = self.arg.getPhysicalIndices(start, stop)
        
        physicalText = str(self.arg)
        _stop = stop
        #ignore white spaces to the end;
        while _stop < len(physicalText) and physicalText[_stop].isspace():
            _stop += 1
        
        if _stop == len(physicalText):
            if not start:
                stop = _stop + self.tagEnd.length

        stop += self.tagBegin.length
        
        if start:
            start += self.tagBegin.length
            
        return start, stop

    def replace(self, old, node):
        assert self.arg is old, self
        self.arg = node
        if node.parent:
            assert node.parent is self
        else:
            node.parent = self

    def enumerate(self):        
        yield from self.ptr.enumerate()

    def has(self, cls):
        if isinstance(self, cls):
            return True
        return self.ptr.has(cls)

    def reduceToNodeText(self):
        arg, tagBegin, tagEnd, parent = self.arg, self.tagBegin, self.tagEnd, self.parent
        assert tagEnd is None, "tagEnd == null"
        if parent.is_XMLNodeArray:
            index = parent.args.index(self)
            count = 1
            
            if arg.is_XMLNodeArray:
                args = arg.args
                for arg in args:
                    arg.parent = parent
                
                if args[0].is_XMLNodeText:
                    args[0].start = tagBegin.start
                else:
                    args.insert(0, XMLNodeText(tagBegin.reduceToNodeText(), parent))
                    
                if index and parent.args[index - 1].is_XMLNodeText:
                    index -= 1
                    count += 1
                    args[0].start = parent.args[index].start
                    
            elif arg.is_XMLNodeCaret:
                arg = XMLNodeText(self.tagBegin.reduceToNodeText(), parent)
                if index + 1 < len(parent.args) and parent.args[index + 1].is_XMLNodeText:
                    count += 1
                    arg.stop = parent.args[index + 1].stop
                
                if index and parent.args[index - 1].is_XMLNodeText:
                    index -= 1
                    count += 1
                    arg.start = parent.args[index].start
                
                args = [arg]
            else:
                arg.parent = parent
                if arg.is_XMLNodeText:
                    if index + 1 > len(parent.args) and parent.args[index + 1].is_XMLNodeText:
                        count += 1
                        arg.stop = parent.args[index + 1].stop
                    
                    if index and parent.args[index - 1].is_XMLNodeText:
                        index -= 1
                        count += 1
                        arg.start = parent.args[index].start
                    else:
                        arg.start = tagBegin.start
                    args = [arg]
                else:
                    args = [XMLNodeText(self.tagBegin.reduceToNodeText(), parent), arg]
                    if index and parent.args[index - 1].is_XMLNodeText:
                        index -= 1
                        count += 1
                        args[0].start = parent.args[index].start
                    
            splice(parent.args, index, count, *args)
                    
        else:
            assert parent.is_XMLNodeBinaryTag
            if arg.is_XMLNodeArray:
                args = arg.args
                if args[0].is_XMLNodeText:
                    args[0].start = tagBegin.start
                else:
                    args.insert(0, XMLNodeText(tagBegin.reduceToNodeText(), arg))
            elif arg.is_XMLNodeText:
                arg.start = tagBegin.start
            elif arg.is_XMLNodeCaret:
                arg = XMLNodeText(tagBegin.reduceToNodeText())
            else:
                arg = XMLNodeArray([XMLNodeText(self.tagBegin.reduceToNodeText()), arg])
                
            arg.parent = parent
            parent.replace(self, arg)
    
    def append_tag(self, node):
        parent = self.parent
        if isinstance(parent, XMLNodeArray):
            node.parent = parent
            assert not node.is_XMLNodeCaret
            parent.args.append(node)
            return node
        else:
            array = XMLNodeArray([self, node], parent)
            if parent:
                parent.replace(self, array)
            return array
        
    
class XMLNodeSingleTag(XMLNode):
    is_XMLNodeSingleTag = True
    
    def __init__(self, arg, parent=None):
        super().__init__(parent)
        self.arg = arg

    @property
    def tag(self):
        return self.arg.tag
    
    @property
    def start(self):
        return self.arg.start
    
    @property
    def stop(self):
        return self.arg.stop
    
    @computed
    def physicalText(self):
        return str(self.arg)

    @computed
    def text(self):
        return self.arg.text

    @property
    def logicalLength(self):
        return 1

    @computed
    def texts(self):
        return [self.text]
    
    @computed
    def style(self):
        _style = {}
        _style[self.tag] = Range(0, len(self.text))
        return _style

    def logical2physical(self, pos):
        return self.arg.length - 2

    def physical2logical(self, pos):
        return 0
    
    def is_physically_valid(self, pos):
        ...
            
    def is_sibling(self, start, end):
        ...
            
    def getPhysicalIndices(self, start, stop):
        return 0, self.arg.length

    def enumerate(self):        
        yield from self.ptr.enumerate()

    def has(self, cls):
        if isinstance(self, cls):
            return True
        return self.ptr.has(cls)
    
    
class XMLNodeArray(XMLNode):
    is_XMLNodeArray = True
    
    def __init__(self, args, parent=None):
        super().__init__(parent)
        self.args = args
        for ptr in args:
            ptr.parent = self

    def append_left_tag(self, tag):
        caret = XMLNodeCaret()
        self.args.append(XMLNodeBinaryTag(tag, caret, parent=self))
        return caret

    def append_text_node(self, text):
        node = XMLNodeText(text, parent=self)
        assert not self.args[-1].is_XMLNodeText
        self.args.append(node)
        return node
    
    def append_single_tag(self, tag):
        node = XMLNodeSingleTag(tag, parent=self)
        self.args.append(node)
        return node
    
    def append_tag(self, node):
        node.parent = self
        assert not node.is_XMLNodeCaret
        self.args.append(node)
        return node
    
    @computed
    def physicalText(self):
        return ''.join(str(nd) for nd in self.args)

    @property
    def start(self):
        return self.args[0].start
    
    @property
    def stop(self):
        return self.args[-1].stop

    @computed
    def text(self):
        return ''.join(nd.text for nd in self.args)
    
    @property
    def logicalLength(self):
        return sum(el.logicalLength for el in self.args)

    @computed
    def texts(self):
        return [text for nd in self.args for text in nd.texts]
    
    @computed
    def zeros(self):
        zeros = []
        length = 0
        for arg in self.args:
            zeros += [zero + length for zero in arg.zeros]
            length += len(arg.text)
        return zeros

    @computed
    def style(self):
        for i, tagEnd in enumerate(self.args):
            if tagEnd.is_XMLNodeUnbalancedTag:
                _self = tagEnd.binaryTag
                assert _self.tagEnd is tagEnd, "_self.tagEnd is tagEnd"
                while True:
                    parent = _self.parent
                    if parent.is_XMLNodeArray:
                        index = parent.args.index(_self)
                        for j in range(index + 1, i if parent is self else len(parent.args)):
                            parent.args[j].modify_style(tagEnd.tag)
                    _self = parent
                    if parent is self:
                        break
        
        _style = {}
        length = 0
        for arg in self.args:
            for tag, set in arg.style.items():
                set += length
                if tag in _style:
                    _style[tag] = _style[tag].union_without_merging(set)
                else:
                    _style[tag] = set
            
            length += len(arg.text)
        
        return _style

    def replace(self, old, node):
        i = self.args.index(old)
        self.args[i] = node
        assert not node.is_XMLNodeArray
        if node.parent:
            assert node.parent is self
        else:
            node.parent = self

    def enumerate(self):        
        for node in self.args:
            yield from node.enumerate()

    def has(self, cls):
        if isinstance(self, cls):
            return True
        for e in self.args:
            if e.has(cls):
                return True
        return False

    def interval(self, className):
        text = self.text
        if text:
            zeros = self.zeros
            if not zeros or zeros[0]:
                zeros.insert(0, 0)
            
            if not zeros or zeros[-1] < len(text):
                zeros.append(len(text))
                
            interval = map(lambda i: Object(offsetStart=zeros[i - 1], offsetStop=zeros[i], className=className), range(1, len(zeros)))
            interval = [*interval]
        else:
            interval = []

        return detect_style(interval, self.style)
    
    @computed
    def logicalOffset(self):
        logicalOffset = []
        start = 0
        for arg in self.args:
            stop = start + arg.logicalLength
            logicalOffset.append((start, stop))
            start = stop
        
        return logicalOffset
    
    @computed
    def physicalOffset(self):
        physicalOffset = []
        start = 0
        for arg in self.args:
            stop = start + arg.length
            physicalOffset.append((start, stop))
            start = stop
            
        return physicalOffset
    
    @computed
    def offsets(self):
        offsets = []
        offset = 0
        for arg in self.args:
            offsets.append(offset)
            offset += arg.length - arg.logicalLength
        
        return offsets
        
    def logical2physical(self, pos):
        logicalOffset, offsets = self.logicalOffset, self.offsets
        assert logicalOffset, logicalOffset
        index = binary_search(logicalOffset, pos, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)

        prev_start = logicalOffset[index][0]
        return self.args[index].logical2physical(pos - prev_start) + prev_start + offsets[index]
    
    def physical2logical(self, pos):
        physicalOffset, offsets = self.physicalOffset, self.offsets
        assert physicalOffset, physicalOffset
        index = binary_search(physicalOffset, pos, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)

        prev_start = physicalOffset[index][0]
        return self.args[index].physical2logical(pos - prev_start) + prev_start - offsets[index]
    
    def is_physically_valid(self, pos):
        physicalOffset, offsets = self.physicalOffset, self.offsets
        assert physicalOffset, physicalOffset
        index = binary_search(physicalOffset, pos, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
        if index < len(physicalOffset):
            prev_start = physicalOffset[index][0]
            pos -= prev_start
            if pos >= 0:
                return self.args[index].is_physically_valid(pos)
        
    def is_sibling(self, start, end):
        if start < end:
            physicalOffset, offsets = self.physicalOffset, self.offsets
            assert physicalOffset, physicalOffset
            index = binary_search(physicalOffset, start, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
            _index = binary_search(physicalOffset, end - 1, lambda args, hit: -1 if hit >= args[1] else 1 if hit < args[0] else 0)
            if _index == index < len(physicalOffset):
                prev_start = physicalOffset[index][0]
                start -= prev_start
                end -= prev_start
                if start >= 0:
                    return self.args[index].is_sibling(start, end)
                
            if index < _index < len(physicalOffset):
                if self.args[index].is_XMLNodeText:
                    if self.args[_index].is_XMLNodeText:
                        return True
                    
                    if self.args[_index].is_XMLNodeBinaryTag or self.args[_index].is_XMLNodeSingleTag:
                        prev_start = physicalOffset[_index][0]
                        end -= prev_start
                        if end == self.args[_index].length:
                            return True

    def cmp(self, args, hit):
        return -1 if hit >= args[1] else 1 if hit < args[0] else 0
        
    def getPhysicalIndices(self, start, stop):
        logicalOffset, offsets = self.logicalOffset, self.offsets
        index = binary_search(logicalOffset, start, self.cmp)
        _index = binary_search(logicalOffset, stop - 1, self.cmp)
        
        if index == _index:
            prev_start = logicalOffset[index][0]    
            return add(self.args[index].getPhysicalIndices(start - prev_start, stop - prev_start), prev_start + offsets[index])
        else:
            [prev_start, prev_stop] = logicalOffset[index]
            _prev_start = logicalOffset[_index][0]    
            start = add(self.args[index].getPhysicalIndices(start - prev_start, prev_stop - prev_start), prev_start + offsets[index])[0]
            stop = add(self.args[_index].getPhysicalIndices(0, stop - _prev_start), _prev_start + offsets[_index])[1]
            return [start, stop]

    def sanctity_check(self):
        src = str(self)
        for args in self.interval(''):
            offsetStart, offsetStop, className = args.offsetStart, args.offsetStop, args.className
            physicalStart = self.logical2physical(offsetStart)
            physicalStop = physicalStart + offsetStop - offsetStart
            if not re.search('img|mspace|br|entity-(#[0-9]+|#x[0-9a-f]+|[^\t\n\f <&#;]{1,32})', className, re.IGNORECASE):
                if src[physicalStart:physicalStop] != self.text[offsetStart:offsetStop]:
                    print(f'{src[physicalStart:physicalStop]} != {self.text[offsetStart:offsetStop]}')
                    physicalStart = self.logical2physical(offsetStart)
                    physicalStop = physicalStart + offsetStop - offsetStart
                    print(physicalStart, physicalStop)
        
        
class XMLNodeUnbalancedTag(XMLNode):
    is_XMLNodeUnbalancedTag = True

    def __init__(self, tagEnd, binaryTag, parent=None):
        super().__init__(parent)
        self.tagEnd = tagEnd
        self.binaryTag = binaryTag
        binaryTag.tagEnd = self

    @property
    def tag(self):
        return self.tagEnd.tag
    
    @property
    def start(self):
        return self.tagEnd.start
    
    @property
    def stop(self):
        return self.tagEnd.stop
    
    @computed
    def physicalText(self):
        return str(self.tagEnd)

    @computed
    def text(self):
        return ''

    @property
    def logicalLength(self):
        return 0

    @computed
    def texts(self):
        return []
    
    @computed
    def zeros(self):
        return [0]
    
        
def compile(infix):
    caret = XMLNodeCaret()

    leftTagCount = {}
    for text in infix:
        if text.is_TagBegin:
            caret = caret.append_left_tag(text)
            if text.tag not in leftTagCount:
                leftTagCount[text.tag] = []
            leftTagCount[text.tag].append(caret.parent)
#            assert caret.parent.tagBegin is text, "caret.parent.tagBegin is text"
            
        elif text.is_TagEnd:
            old = caret
            while True:
                if caret:
                    root = caret
                    caret = caret.parent
                    if isinstance(caret, XMLNodeBinaryTag) and caret.tag == text.tag:
#                         assert caret.tagEnd is None
                        caret.set_tagEnd(text)
                        while True:
                            _caret = leftTagCount[text.tag].pop()
                            if caret is _caret:
                                break
                            _caret.reduceToNodeText()
#                         index = leftTagCount[text.tag].index(caret)
#                         leftTagCount[text.tag].pop(index)
                        break
                elif leftTagCount[text.tag]:
                    parent = leftTagCount[text.tag].pop()
                    caret = XMLNodeUnbalancedTag(text, parent)
                    while parent.stop < root.stop:
                        parent = parent.parent
                        
                    caret = parent.append_tag(caret)
                    break
                else:
                    text = text.reduceToNodeText()
                    caret = old.append_text_node(text)
                    break
                
        elif text.is_TagSingle or text.is_HTMLEntity:
            caret = caret.append_single_tag(text)
        else:
#             assert text.is_PlainText
            caret = caret.append_text_node(text)
        
    for tag in leftTagCount:
        while leftTagCount[tag]:
            leftTagCount[tag].pop().reduceToNodeText()

    return caret.root
    
    
def style_type(ptr, style):
    offsetStart = ptr.offsetStart
    offsetStop = ptr.offsetStop        
    this_set = Range(offsetStart, offsetStop)
    
    style_intersected = {}
    for tag in style:
        set = style[tag]
        if isinstance(set, dict):
            set = Union(**set) if 'args' in set else Range(**set)
        
        intersection = set & this_set
        
        if not intersection:
            continue
        
        style_intersected[tag] = intersection

    if not style_intersected:
        return
    
    indicator = [Object(className=ptr.className) for i in range(offsetStop - offsetStart)]
    
    def processRangeObject(set, tag):
        start = set.start
        stop = set.stop
        for i in range(start, stop):
            className = indicator[i].className
            if className:
                indicator[i].className += ' ' + tag
            else:
                indicator[i].className = tag
    
    for tag in style_intersected:
        set = style_intersected[tag]
        set = set.__add__(-offsetStart)
            
        args = [set] if set.is_Range else set.args
        for i, s in enumerate(args):
            if i and args[i - 1].stop == s.start:
                tag = toggleCase(tag)
            processRangeObject(s, tag)

    interval = []
    i = 0
    while i < len(indicator):
        j = i + 1
        while j < len(indicator):
            if indicator[j].className != indicator[i].className:
                break
            j += 1
        
        className = indicator[i].className
        if not className:
            className = ptr.className
            
        interval.append(Object(offsetStart=i + offsetStart, offsetStop=j + offsetStart, className=className))
        i = j
        
    return interval
    
    
def detect_style(interval, style):
    i = 0
    while i < len(interval):
        ptr = interval[i]
        
        stype = style_type(ptr, style)
        if isinstance(stype, list):
            interval = interval[:i] + stype + interval[i + 1:]
            i += len(stype) - 1
        elif stype:
            ptr.className += f" {style}"
        
        i += 1
            
    return interval

tagName = '([a-z][-:_a-z]*\d*)'
attrList = '''(?:\s+:?[a-z][-:_a-z]*=(?:"[^"]*"|'[^']*'))*\s*'''    
tagBegin = f'<{tagName}{attrList}>'
tagRegex = fr'''{tagBegin}(?=[\s\S]*?<\/\1>)|</{tagName}>|<(img|mspace|br){attrList}/>|&(#[0-9]+|#x[0-9a-f]+|[^\t\n\f <&#;]{{1,32}});'''

# print("tagRegex =", tagRegex)

def construct_rich_text(text):
    start = 0
    
    richTexts = []
    leftTagCount = {} 
    for m in re.finditer(tagRegex, text, re.IGNORECASE):
#                             1-----------------)                                                                     2-----------------)   3-------------)                                                   4----------------------------------------)
        prevText = text[start:m.start()]
        if prevText:
            richTexts.append(PlainText(text, start, m.start()))

        if m[1]:
            tag = m[1]
            richText = TagBegin(text, m.start(), m.end(), tag)
            if tag not in leftTagCount:
                leftTagCount[tag] = 0
            leftTagCount[tag] += 1
        elif m[2]:
            tag = m[2]
            if leftTagCount.get(tag):
                leftTagCount[tag] -= 1
                richText = TagEnd(text, m.start(), m.end(), tag)
            elif prevText:
                richText = richTexts.pop()
                richText.stop = m.end()
            else:
                richText = PlainText(text, m.start(), m.end())
        elif m[3]:
            richText = TagSingle(text, m.start(), m.end(), m[3])
        else:
            richText = HTMLEntity(text, m.start(), m.end(), m[4])
        
        richTexts.append(richText)
        
        start = m.end()
    
    restText = text[start:]
    if restText:
        richTexts.append(PlainText(text, start, len(text)))

    return compile(richTexts)


class XMLText:
    def __init__(self, src, start, stop):
        self.src = src
        self.start = start
        self.stop = stop
    
    def __str__(self):
        return self.src[self.start:self.stop]
    
    __repr__ = __str__
    
    is_PlainText = False
    is_TagEnd = False
    is_TagBegin = False
    is_TagSingle = False
    is_HTMLEntity = False
    
    @property
    def length(self):
        return self.stop - self.start
    
    def reduceToNodeText(self):
        return PlainText(self.src, self.start, self.stop)


class TagBegin(XMLText):
    is_TagBegin = True
    
    def __init__(self, src, start, stop, tag):
        super().__init__(src, start, stop)
        self.tag = tag
        
    
class TagEnd(XMLText):
    is_TagEnd = True
    
    def __init__(self, src, start, stop, tag):
        super().__init__(src, start, stop)
        self.tag = tag


class TagSingle(XMLText):
    is_TagSingle = True
    
    def __init__(self, src, start, stop, tag):
        super().__init__(src, start, stop)
        self.tag = tag
        
        text = tag.lower()
        if tag == 'img':
            text = '★'
        elif tag == 'mspace':
            text = ' '
        elif tag == 'br':
            text = '\n'
        else:
            text = '?'
            
        self.text = text


class HTMLEntity(XMLText):
    is_HTMLEntity = True
    
    def __init__(self, src, start, stop, tag):
        super().__init__(src, start, stop)
        self.tag = 'entity-' + tag
        text = html.unescape(f'&{tag};')
        if len(text) != 1:
            text = '?'
        
        self.text = text
        
    
class PlainText(XMLText):
    is_PlainText = True
    
    def __init__(self, src, start, stop):
        super().__init__(src, start, stop)

    @property
    def text(self):
        return str(self)
    

def cmp_range(a, b):
    if a.stop <= b.start:
        return -1
            
    if a.start >= b.stop:
        return 1
    
    return 0
    
    
def is_alphanum(ch):    
    return re.compile('[\p{Letter}\d]').match(ch)


def convertToOriginal(arr):
    s = ""
    for i in range(len(arr) - 1):
        s += arr[i]

        if is_alphanum(arr[i][-1]) and is_alphanum(arr[i + 1][0]):
            s += " "

    s += arr[-1]
    return s

def is_punct(ch):
    return re.compile('\p{Punctuation}').match(ch)

def convertToOriginalWithOffsets(arr):
    s = ""
    start = 0
    offsets = []
    for i in range(len(arr) - 1):
        stop = start + len(arr[i])
        offsets.append((start, stop))
        s += arr[i]

        if (is_alphanum(arr[i][-1]) or is_punct(arr[i][-1])) and is_alphanum(arr[i + 1][0]):
            s += " "
            stop += 1
            
        start = stop

    stop = start + len(arr[-1])
    offsets.append((start, stop))
    s += arr[-1]
    
    return s, offsets

def test():
    physicalText = "firstly,&nbsp;<div><b>this&nbsp;is&nbsp;<I>an&nbsp;<font color='red'><i>italic</i>&nbsp;<u>&lt;underline&gt;</u><mspace /></font>, simply put</b>,&nbsp;this&nbsp;is&nbsp;an&nbsp;italic&nbsp;text,&nbsp;</I>with&nbsp;something&emsp;<u>&lt;underlined&gt;</u>;</div>&ensp;<b>another&nbsp;bold&nbsp;text&nbsp;is&nbsp;followed.&nbsp;</b>At&nbsp;last,&nbsp;there&nbsp;a&nbsp;plain&nbsp;text."
    richText = construct_rich_text(physicalText)
    # print(richText)
    richText.sanctity_check()
    print("physical text =", richText.physicalText)
    print(" logical text =", richText.text)
    print(" style_traits =", richText.style_traits)
    
    # for tag, set in richText.style.items():
    #     if set.is_Range:
    #         print(tag, ":", richText.text[slice(*set.args)])
    #     else:
    #         print(tag, ":", '\t'.join(map(lambda arg: richText.text[slice(*arg.args)], set.args)))
            
    start, stop = 20, 39
    print("the logical text is :")
    print(richText.text[start:stop])
    print("its style trait is :")
    print(richText.style_traits[start:stop])
    
    start, stop = richText.getPhysicalIndices(start, stop)
    print("its original physical text is:")
    print(physicalText[start:stop])
    print()
    
if __name__ == '__main__':
    test()


def HeapPermutation(k, A):
    '''
    https://en.wikipedia.org/wiki/Heap%27s_algorithm
    '''
    
    if k == 1:
        yield A
    else:
        # Generate permutations with kth unaltered, Initially k == length(A)
        yield from HeapPermutation(k - 1, A)
        
        # Generate permutations for kth swapped with each k-1 initial
        for i in range(k - 1):
            # Swap choice dependent on parity of k (even or odd)
            if k & 1:
                A[0], A[k - 1] = A[k - 1], A[0]
            else:
                A[i], A[k - 1] = A[k - 1], A[i]
            
            yield from HeapPermutation(k - 1, A)


def generate_all_permutation(A):
    yield from HeapPermutation(len(A), A)        

        
def skip_first_permutation(A):
    generator = HeapPermutation(len(A), A)
    next(generator)
    yield from generator


import heapq


class TopKHeap:

    def __init__(self, k):
        self.data = []
        self.k = k

    def __str__(self):
        return str(self.data)
    
    __repr__ = __str__
    
    def push(self, num):
        #internally, it is a minimum heap!
        if len(self.data) < self.k:
            heapq.heappush(self.data, num)
        elif num > self.data[0]:
            heapq.heapreplace(self.data, num)

    def topk(self):
        
        arr = []
        while self.data: 
            a = heapq.heappop(self.data)
            arr.append(a)
            
        arr.reverse()
        return arr
    




# -*- coding: UTF-8 -*-
from intervaltree import IntervalTree


class Entity2Label:
    def __init__(self, label_schema='BILOU', resolve_conflict=True):
        """
        entity to label transformer
        :param label_schema: label schema, only allow 'BILOU', 'BIO' and None.
                            When its value is none, label will be uppercase of entity type(for semi-markov CRF)
        :param resolve_conflict: whether resolve conflict which entity start or end is not in token starts and ends list
        """
        if label_schema not in {'BILOU', 'BIO', None}:
            raise Exception(label_schema)
        self.label_schema = label_schema
        self.resolve_conflict = resolve_conflict

    def sent_batch(self, sent_list):
        """
        transform entities and text in paragraph list to labels
        :param sent_list: paragraph list
        :return:
        """
        for item in sent_list:
            self.sent(item)
        return sent_list

    def sent(self, sent):
        """
        transform text, entities, token and pos tags into labels, labels will attach in data object
        For example:
            input: sent = {'text': 'Steve Jobs founded Apple.Inc.',
                'tokens': [{'end': 5, 'start': 0, 'text': 'Steve'},
                  {'end': 10, 'start': 6, 'text': 'Jobs'},
                  {'end': 18, 'start': 11, 'text': 'founded'},
                  {'end': 24, 'start': 19, 'text': 'Apple'},
                  {'end': 25, 'start': 24, 'text': '.'},
                  {'end': 28, 'start': 25, 'text': 'Inc'},
                  {'end': 29, 'start': 28, 'text': '.'}],
                  'entities': [{'end': 10, 'entity': 'Steve Jobs', 'start': 0, 'type': 'PER'},
                         {'end': 28, 'entity': 'Apple.Inc', 'start': 19, 'type': 'COMP'}]
                  }
            output: ['B-PER', 'L-PER', 'O', 'B-COMP', 'I-COMP', 'L-COMP', 'O']
        :param sent: sentence to be transformed, must be dict type, and has text, entities and tokens key.
        :return: new data dict with labels
        """
        if not isinstance(sent, dict):
            raise TypeError('sent must be in dict type')
        if 'text' not in sent:
            raise KeyError('must have text field')
        elif 'entities' not in sent:
            raise KeyError('muse have entities field')
        elif 'tokens' not in sent:
            raise KeyError('must have tokens field')

        text = sent['text']
        entities = sent['entities']
        tokens = sent['tokens']

        if self.resolve_conflict:
            tokens = split_conflict_spans(entities, tokens, text)

        token_spans = sorted([(t['start'], t['end']) for t in tokens])
        labels = ['O'] * len(token_spans)

        for entity in entities:
            start_index, end_index = get_index_span(token_spans,
                                                    entity['start'],
                                                    entity['end'],
                                                    is_sorted=True)
            entity_labels = self.single(entity, start_index, end_index)
            labels[start_index:end_index] = entity_labels

        sent['tokens'] = tokens
        sent['labels'] = labels

        return sent

    def single(self, entity, start_index, end_index, type_in_label=None, label_suffix=None):
        """
        transform single entity to labels
        :param entity: entity in dict. must have entity, start, end and type key
        :param start_index: entity first token index in token list
        :param end_index: entity last token index in token list
        :param type_in_label: entity type string in label schema, default is None, assigned with entity type 
        :param label_suffix: suffix string in label
        :return: labels represent entity
        """
        if type_in_label is None:
            type_in_label = entity['type']

        if not type_in_label:
            raise ValueError('entity type is empty')

        start_label = join_label('B', type_in_label, label_suffix)
        inter_label = join_label('I', type_in_label, label_suffix)
        end_label = join_label('L', type_in_label, label_suffix)
        unique_label = join_label('U', type_in_label, label_suffix)
        if self.label_schema == 'BILOU':
            if end_index - start_index == 1:
                entity_labels = [unique_label]
            else:
                entity_labels = [start_label]
                if end_index - start_index > 2:
                    entity_labels.extend([inter_label] * (end_index - start_index - 2))
                entity_labels.append(end_label)
        elif self.label_schema == 'BIO':
            entity_labels = [start_label]
            if end_index - start_index > 1:
                entity_labels.extend([inter_label] * (end_index - start_index - 1))
        elif not self.label_schema:
            entity_labels = [entity['type'].upper()] * (end_index - start_index)
        else:
            raise Exception('label schema is not supported.')
        return entity_labels


class EntityRelation2Label(Entity2Label):
    """
    transform entities and relations into labels with NovelTag schema in papar
    Zheng, Suncong, et al. "Joint extraction of entities and relations based on a novel tagging scheme." arXiv preprint arXiv:1706.05075 (2017).
    """

    def __init__(self):
        super().__init__()

    def sent(self, sent):
        if not isinstance(sent, dict):
            raise TypeError('sent must be in dict type')
        if 'text' not in sent:
            raise KeyError('must have text field')
        elif 'entities' not in sent:
            raise KeyError('muse have entities field')
        elif 'tokens' not in sent:
            raise KeyError('must have tokens field')
        elif 'relations' not in sent:
            raise KeyError('must have relations field')

        for e in sent['entities']:
            if 'id' not in e:
                raise KeyError('must have id field in entity for building relation')

        text = sent['text']
        entities = sent['entities']
        id2entity = {e['id']: e for e in entities}
        if len(entities) > len(id2entity):
            raise ValueError('entity ids have duplication! ')

        tokens = sent['tokens']
        relations = sent['relations']

        if self.resolve_conflict:
            tokens = split_conflict_spans(entities, tokens, text)

        token_spans = sorted([(t['start'], t['end']) for t in tokens])
        labels = ['O'] * len(token_spans)
        for relation in relations:
            first_entity_id = relation['e1']
            second_entity_id = relation['e2']
            rel_type = relation['type']
            labels = self.assign_labels_with_entity(token_spans, id2entity[first_entity_id],
                                                    labels, rel_type, 1)
            labels = self.assign_labels_with_entity(token_spans, id2entity[second_entity_id],
                                                    labels, rel_type, 2)

    def assign_labels_with_entity(self, token_spans, entity, labels, rel_type, entity_idx):
        start_index, end_index = get_index_span(token_spans,
                                                entity['start'],
                                                entity['end'],
                                                is_sorted=True)
        entity_labels = self.single(entity, start_index, end_index, rel_type, entity_idx)
        labels[start_index:end_index] = entity_labels
        return labels


def get_index_span(spans, start, end, is_sorted=False):
    """
    get start and end of index position in spans by start and end value in spans
    start is first value of tuple item in spans,
    end is second value of tuple item in spans.
    start and end must occur in spans

    For example,
        input spans = [(1, 5), (5, 10), (10,25), (25, 100)], start = 5, end = 100
        output: (1, 4)

    :param spans: span list to do search, every item is a two elements tuple.
    :param start: start position to be searched in  span starts
    :param end: end position to be searched in span ends
    :param is_sorted: Whether spans have been sorted in ascend order, default is False.
                      When you confirm that, set it True to save time.
    :return: start and end index in spans
    """
    if not is_sorted:
        spans = sorted(spans)

    start_index = end_index = -1
    # indicators to check span duplicates
    last_start = last_end = -1
    for idx, (span_start, span_end) in enumerate(spans):
        if last_start == span_start:
            raise ValueError('spans has duplicated start {}'.format(span_start))
        if last_end == span_end:
            raise ValueError('spans has duplicated end {}'.format(span_end))

        if span_start == start:
            start_index = idx
        if span_end == end:
            end_index = idx + 1
            break

        last_start = span_start
        last_end = span_end

    if start_index == -1:
        raise ValueError('start not in span starts')
    if end_index == -1:
        raise ValueError('end not in span ends')
    return start_index, end_index


def join_label(label, entity_type, suffix=None):
    if suffix is None:
        final_label = label + '_' + entity_type
    else:
        final_label = label + '_' + entity_type + '_' + str(suffix)
    return final_label


def split_conflict_spans(entities, tokens, text):
    """
    resolve the conflict in tokens and entities
    :param entities: entity list
    :param tokens: token list, everyone is a dict which has text, start, end and pos tag (optional)
    :param text: original text
    :return: new tokens whose conflicts have been resolved.
    """
    if not tokens:
        return tokens
    first_token_start = tokens[0]['start']
    last_token_end = tokens[-1]['end']

    token_spans = [(t['start'], t['end']) for t in tokens]
    token_starts, token_ends = list(zip(*token_spans))
    split_indices = []

    # get conflict index in sentence,
    # conflict means entity start not in token starts or entity end not in token ends
    for e in entities:
        if e['start'] < first_token_start:
            cross_boundary_msg = 'entity start {} is less than first token start {}'
            raise ValueError(cross_boundary_msg.format(e['start'], first_token_start))
        elif e['end'] > last_token_end:
            cross_boundary_msg = 'entity end {} is larger than last token end {}'
            raise ValueError(cross_boundary_msg.format(e['end'], last_token_end))
        if e['start'] not in token_starts:
            split_indices.append(e['start'])
        if e['end'] not in token_ends:
            split_indices.append(e['end'])
    split_indices = sorted(set(split_indices))

    # split token cross conflict index into two tokens
    for split_index in split_indices:
        tree = IntervalTree.from_tuples(token_spans)
        interval = list(tree.at(split_index))
        if len(interval) != 1:
            raise Exception('interval count error')
        interval = interval[0]
        if interval.begin == split_index:
            raise Exception('split index start error')
        elif interval.end == split_index:
            raise Exception('split index end error')
        token_spans.extend([(interval.begin, split_index), (split_index, interval.end)])
        if (interval.begin, interval.end) in token_spans:
            token_spans.remove((interval.begin, interval.end))

    token_spans = sorted(token_spans, key=lambda s: s[0])
    new_tokens = [{'text': text[s:e], 'start': s, 'end': e} for (s, e) in token_spans]

    # resolve conflict in pos_tag such as token
    # conflict tokens
    # todo: pos tag conflict resolving with some rules
    if 'pos_tag' in tokens[0]:
        pos_tags = []
        token_starts, token_ends = list(zip(*token_spans))
        for token in tokens:
            pos_tag = token['pos_tag']
            start = token['start']
            end = token['end']
            pos_start = token_starts.index(start)
            pos_end = token_ends.index(end)
            if pos_start == pos_end:
                pos_tags.append(pos_tag)
            elif pos_end - pos_start > 0:
                pos_tags.extend([pos_tag] * (pos_end - pos_start + 1))
            else:
                raise ValueError('token end index is before token start index')
        pos_tag_count = len(pos_tags)
        new_token_count = len(new_tokens)
        msg = 'pos tag count and token text count are not equal, token {}, pos tag {}'
        if pos_tag_count != new_token_count:
            raise Exception(msg.format(new_token_count, pos_tag_count))
        else:
            for token, pos_tag in zip(new_tokens, pos_tags):
                token['pos_tag'] = pos_tag
    return new_tokens



import json
import requests
from json.decoder import JSONDecodeError

def json_post(url, data):
    try:
        return json.loads(
            requests.post(
                url,
                data=json.dumps(data, ensure_ascii=False).encode(encoding='utf-8'), 
                headers={"Content-Type": "application/json"}).text)
    except Exception as e:
        return e


def json_post_curl(url, data):
    data = json.dumps(data, ensure_ascii=False)#.encode(encoding='utf-8')
    return f"curl -H 'Content-Type: application/json' -X POST '{url}' -d'{data}'"
    
def get(url, **kwargs):
    import regex as re
    if kwargs:
        url += "?"
        parameters = []
        for key, value in kwargs.items():
            if isinstance(value, str):
                value = re.sub(r'\+', '%2B', value)

            parameters.append(f"{key}={value}")
        url += '&'.join(parameters)

    text = requests.get(url).text
    try:
        return json.loads(text)
    except JSONDecodeError:
        return text
    
def form_post(url, data):
    for key, value in data.items():
        if not isinstance(value, str):
            data[key] = json.dumps(value, ensure_ascii=False)
            
    text = requests.post(
        url,
        data=data, 
        headers={"Content-Type": "application/x-www-form-urlencoded; charset=UTF-8"}).text
    try:
        return json.loads(text)
    except JSONDecodeError:
        return text


def octet_stream_post(url, filename, data):
    text = requests.post(
        url,
        data=data,
        headers={"Content-Type": "application/octet-stream", 'filename': filename, 'Content-Length': str(len(data))}).text
    try:
        return json.loads(text)
    except JSONDecodeError:
        return text



from setuptools import setup

if __name__ == "__main__":
    setup(package_data={'': ['lib/eigen.dll']})
    
# usage of twine:
# pip install twine
# python setup.py install
# del dist\*.tar.gz && python setup.py sdist && twine upload dist/*.tar.gz 
# del dist\*.tar.gz && python setup.py sdist && twine upload -r nexus dist/*.tar.gz
# pip uninstall -y std.algorithm && pip install --upgrade std.algorithm
# pip install --upgrade std.algorithm -i https://pypi.Python.org/simple/


import os
import json

def createNewFile(path):
    createNewPath(os.path.dirname(path))

    try:
        os.mknod(path)
    except:
        with open(path, 'wb') as _:
            ...


def createNewPath(basedir):
    if not os.path.exists(basedir):
        os.makedirs(basedir)


class Text:

    def __init__(self, path, encoding="utf-8", **kwargs):
        try:
            self.file = open(path, "r+" if os.path.isfile(path) else "w+", encoding=encoding)
        except PermissionError: 
            self.file = open(path, "r", encoding=encoding)
        except FileNotFoundError:
            createNewFile(path)
            self.file = open(path, "w+", encoding=encoding)

        self.rewind()
        
        if kwargs.get('strip'):
            self.rstrip = self.lstrip = True
        else:
            self.rstrip = kwargs.get('rstrip', True)
            self.lstrip = kwargs.get('lstrip', False)

    def __del__(self):
        self.close()

    def close(self):
        self.file.close()
        
    def __iter__(self):  # Get iterator object on iter
        return self

    def __next__(self):
        while True:
            line = self.file.readline()
            if not line:
                self.rewind()
                raise StopIteration
            
            if self.lstrip:
                line = line.lstrip()
                            
            if self.rstrip:
                line = line.rstrip()
            else:
                if line.endswith('\n'):
                    line = line[:-1]
                
            if not self.lstrip or not self.rstrip or line:
                return line

    def prepend(self, s):
        '''
        prepend = append before the list;
        '''
        self.rewind()

        content = self.file.read()
        self.write(s)
#         self.file.seek(0, 2)
        self.file.write(content)
        self.file.flush()

    def append(self, s, end_of_line='\n'):
        self.end()

        if type(s) == str:
            self.file.write(s + end_of_line)
        else:
            for s in s:
                self.file.write(s + end_of_line)
        self.file.flush()

    def insert(self, index, s):
        if index < 0:
            self.end()
            index = -index - 1
            offset = self.file.tell() - 1
            while index > 0:
                index -= 1
                offset -= 1
                while True:
                    offset -= 1
                    _offset = self.file.seek(offset, os.SEEK_SET)
                    assert _offset == offset
                    char = self.file.read(1)
#                     print('char =', char)
                    if char[0] == '\n' or char[0] == '\r':
                        break
        else:
            self.rewind()
            offset = self.f.tell()
            while index > 0:
                index -= 1
                while True:
                    offset += 1
                    _offset = self.f.seek(offset, os.SEEK_SET)
                    assert _offset == offset
                    char = self.f.read(1)
#                     print('char =', char)
                    if char[0] == '\n' or char[0] == '\r':
                        break

        current_pos = self.file.tell()
        assert current_pos == offset + 1

        rest = self.file.read()
        self.file.seek(current_pos, os.SEEK_SET)

        if type(s) == str:
            self.file.write(s + '\n')
        else:
            for s in s:
                self.file.write(s + '\n')

        self.file.write(rest)

        self.file.flush()

    def __getitem__(self, index):
        if index < 0:
            self.end()
            index = -index
            offset = self.f.tell() - 1
            while index > 0:
                index -= 1
                offset -= 1
                while True:
                    offset -= 1
                    _offset = self.f.seek(offset, os.SEEK_SET)
                    assert _offset == offset
                    char = self.f.read(1)
    #                     print('char =', char)
                    if char == '\n' or char == '\r':
                        break
        else:
            self.rewind()
            offset = self.f.tell()
            while index > 0:
                index -= 1
                while True:
                    offset += 1
                    _offset = self.f.seek(offset, os.SEEK_SET)
                    assert _offset == offset
                    char = self.f.read(1)
#                     print('char =', char)
                    if char == '\n' or char == '\r':
                        break
        current_pos = self.f.tell()

        self.f.seek(current_pos, os.SEEK_SET)

        return self.f.readline().strip()

    def write(self, s):
        self.rewind()

        if type(s) == str:
            self.file.write(s + '\n')
        else:
            for s in s:
                self.file.write(str(s) + '\n')
        self.file.truncate()
        self.file.flush()

    def clear(self):
        self.rewind()
        self.file.write('')
        self.file.truncate()
        self.file.flush()

    def rewind(self):
#         skip_byte_order_mark
        self.file.seek(0, os.SEEK_SET)
        byteOrderMark = self.file.read(1)
        if byteOrderMark and ord(byteOrderMark) != 0xFEFF:
            self.file.seek(0, os.SEEK_SET)        

    def end(self):
        self.file.seek(0, os.SEEK_END)

    def collect(self):
        self.rewind()

        arr = []
        for line in self:
            arr.append(line)
        return arr

    def remove(self, value=None, index=None):
        arr = self.collect()
        if index is None:
            if value in arr:
                arr.remove(value)
                self.write(arr)
        else:
            del arr[index]
            self.write(arr)

    def removeDuplicate(self):
        arr = []
        st = set()
        for s in self.collect():
            if s not in st:
                st.add(s)
                arr += [s]

        self.write(arr)

    def read(self):
        return self.file.read()
    
    def tell(self):
        return self.file.tell()
    
    @property
    def size(self):
        tell = self.tell()
        self.end()
        size = self.file.tell()
        self.file.seek(tell, os.SEEK_SET)        
        return size
        
    def endswith(self, end):
        self.end()
        size = len(end)
        offset = self.file.tell() - size
        if offset < 0:
            return False
        
        _offset = self.file.seek(offset, os.SEEK_SET)
        assert _offset == offset
        char = self.file.read(size)        
        return char == end

    def readlines(self): 
        return [s for s in self]
    
    def readline(self):
        try: 
            return next(self)
        except StopIteration:
            return ''
    
    def writelines(self, array): 
        self.rewind()

        for s in array:
            self.file.write(str(s) + '\n')

        self.file.truncate()
        self.file.flush()

    def find(self, regex, bool=True):
        import re
        for i, line in enumerate(self):
            if re.match(regex, line):
                if bool:
                    return True
                return i
        if bool:
            return False
        else:
            return -1


class GzipFile:
    def __init__(self, path, **kwargs):
        import gzip
        try:
            self.file = gzip.open(path, "rb" if os.path.isfile(path) else "wb")
        except PermissionError: 
            self.file = gzip.open(path, "rb")
        except FileNotFoundError:
            createNewFile(path)
            self.file = gzip.open(path, "wb")
            
        self.json = kwargs.get('json', True)

    def __del__(self):
        self.close()

    def close(self):
        self.file.close()
        
    def __iter__(self):  # Get iterator object on iter
        return self

    def __next__(self):
        if line := self.file.readline():
            return json.loads(line) if self.json else line

        raise StopIteration

    def append(self, s):
        if not isinstance(s, str):
            s = json.dumps(s)
            
        self.file.write(s.encode())
        self.file.write(b'\n')

    def readlines(self): 
        return self.file.readlines()
    
    def readline(self):
        try: 
            return next(self)
        except StopIteration:
            return ''    




# https://www.inf.fh-flensburg.de/lang/algorithmen/pattern/sundayen.htm
# https://en.wikipedia.org/wiki/Boyer%E2%80%93Moore_string-search_algorithm
# https://www.jianshu.com/p/2594d312cefd
def sunday(haystack, needle):
    
    needelLength = len(needle)
    
    haystackLength = len(haystack)
    
    dic = {v:needelLength - k for k, v in enumerate(needle)}
    
    end = needelLength
    
    while end <= haystackLength:
        begin = end - needelLength
        if haystack[begin:end] == needle:
            return begin

        if end >= haystackLength:
            return -1

        offset = dic.get(haystack[end])
        if not offset:
            offset = needelLength + 1
        
        end += offset

    return -1
    



import os
import json
import time
import queue
from _collections import defaultdict
import inspect
import functools
import sys
from enum import Enum, unique
import signal


def is_Windows():
    return os.sep == '\\'

    
# is Linux System
def is_Linux():
    return os.sep == '/'


def argmax(arr):
    m = arr[0]
    index = 0
    for i in range(1, len(arr)):
        a = arr[i]
        if a > m:
            index = i
            m = a
        
    return index

def argmin(arr):
    m = arr[0]
    index = 0
    for i in range(1, len(arr)):
        a = arr[i]
        if a < m:
            index = i
            m = a
        
    return index

def rindex(arr, e):
    for i, _e in enumerate(reversed(arr)):
        if e == _e:
            return -i - 1
    else:
        raise ValueError(f'{e} is not in list')


def batch_map(proc, items, processes=8):
    from multiprocessing import Pool
    with Pool(processes=processes) as pool:
        return pool.map(proc, items)    
    

def listdir(rootdir, sufix='.txt', recursive=False):
    for name in os.listdir(rootdir):
        path = os.path.join(rootdir, name)

        if os.path.isdir(path):
            if recursive:
                yield from listdir(path, sufix, recursive=True)
        else:
            if sufix:
                if path.lower().endswith(sufix):
                    yield path
            else:
                yield path

def listfolders(rootdir):
    for name in os.listdir(rootdir):
        path = os.path.join(rootdir, name)
        if os.path.isdir(path):
            yield name


def listdir_recursive(rootdir, sufix='.txt'):
    for name in os.listdir(rootdir):
        path = os.path.join(rootdir, name)

        if os.path.isdir(path):
            yield from listdir_recursive(path, sufix)
        else:
            if sufix:
                if path.endswith(sufix):
                    yield path
            else:
                yield path


def eol_convert(fileName):
    def eol_convert(fileName):
        with open(fileName, "rb") as f:
            data = bytearray(os.path.getsize(fileName))
            f.readinto(data)
            # print(data)
            data = data.replace(b"\r\n", b"\n")
    
        with open(fileName, "wb") as f:
            # print(data)
            f.write(data)    

    if fileName.startswith('.'):
        for file in listdir('.', fileName, recursive=True):
            eol_convert(file)
    else:
        eol_convert(fileName)

def cstring(s):
    return bytes(s, 'utf8')


def json_encode(data, utf8=False, indent=None):
    from types import FunctionType
    if isinstance(data, FunctionType):

        def func(*args, **kwargs): 
            return json_encode(data(*args, **kwargs), utf8=utf8, indent=indent)
        
        func.__name__ = data.__name__
        return func 
    
    s = json.dumps(data, ensure_ascii=False, indent=indent, cls=JSONEncoder)
    if utf8:
        s = s.encode(encoding='utf-8')
    return s 


class Object:

    def __init__(self, *args, **kwargs):
        if args:
            kwargs, = args
            
        for key, value in kwargs.items():
            if isinstance(value, dict):
                value = Object(value)
            
            self.__dict__[key] = value

    def __iter__(self):  # Get iterator object on iter
        return iter(self.__dict__.keys())

    def __getitem__(self, index):
        return self.__dict__.get(index, None)

    def __setitem__(self, index, rhs):
        self.__dict__[index] = rhs
    
    def __delitem__(self, index):
        del self.__dict__[index]
        
    def __setattr__(self, index, rhs): 
        self.__dict__[index] = rhs
        
    def __getattr__(self, index): 
        return self.__dict__.get(index, None)

    def keys(self):
        return self.__dict__.keys()
    
    def values(self):
        return self.__dict__.values()
    
    def items(self):
        return self.__dict__.items()
    
    def serialize(self):
        res = {}
        for k, v in self.__dict__.items():
            if isinstance(v, Object):
                v = v.serialize()
            res[k] = v
        return res
    
    # perform el in self
    def __contains__(self, el):
        return el in self.__dict__
    
    def pop(self, key):
        return self.__dict__.pop(key)
    
    def __str__(self):
        return str(self.__dict__)
    
    __repr__ = __str__
    
    # self | rhs
    # self |= rhs
    def __or__(self, rhs):
        obj = Object()
        for key, value in self.items():
            obj[key] = value

        for key, value in rhs.items():
            obj[key] = value
            
        return obj

    def __len__(self):
        return len(self.__dict__)
    
    def __eq__(self, rhs):
        if len(self) != len(rhs):
            return False
        
        for k, v in self.items():
            if v != rhs[k]:
                return False
        return True

    @staticmethod
    def from_dict(kwargs):
        obj = Object()
        for key, value in kwargs.items():
            if isinstance(value, dict):
                value = Object.from_dict(value)
                
            obj[key] = value
        return obj
      
class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
      
        if isinstance(obj, (bytes, bytearray)):
            return str(obj, encoding='utf-8')
        
        if isinstance(obj, Object): 
            return obj.__dict__
        
        from inspect import isgenerator#, isgeneratorfunction
        if isgenerator(obj):
            return [*obj]
        
        return super().default(self, obj)


class Array(list):
    
    def __init__(self, *args):
        super(Array, self).__init__(*args)

    def __getitem__(self, index):
        if index < len(self):
            return super(Array, self).__getitem__(index)
    
    def __setitem__(self, index, value):
        if index >= len(self):
            self += [None] * (index - len(self) + 1)
            
        super(Array, self).__setitem__(index, value)
    
    def binary_search(self, value, compareTo=None):
        return binary_search(self, value, compareTo)
    
    def equal_range(self, value, compareTo=None):
        return equal_range(self, value, compareTo)
    
    @property
    def length(self):
        return len(self)
    
    
def binary_insert(arr, value, **kwargs):
    index = binary_search(arr, value, **kwargs)
    arr.insert(index, value)  
    
def compareTo(lhs, rhs):
    if isinstance(lhs, str):
        return compareTo([ord(c) for c in lhs], [ord(c) for c in rhs])
    
    if isinstance(lhs, (tuple, list)):
        for lhs, rhs in zip(lhs, rhs):
            cmp = compareTo(lhs, rhs)
            if cmp:
                return cmp
            
        return 0
            
    return lhs - rhs
    
def binary_search(arr, value, compareTo=None, key=None):    
    if compareTo is None:
        if key is None:
            compareTo = lambda x, y: -1 if x < y else 1 if x > y else 0
        else:
            def compareTo(lhs, rhs):
                global compareTo
                return compareTo(key(lhs), key(rhs))
        
    begin = 0
    end = len(arr)
    while True:
        if begin == end:
            return begin
            
        mid = begin + end >> 1
        ret = compareTo(arr[mid], value)
        if ret < 0:
            begin = mid + 1
        elif ret > 0:
            end = mid
        else:
            return mid


def equal_range(arr, value, compareTo=None):
    if compareTo is None:
        compareTo = lambda x, y: -1 if x < y else 1 if x > y else 0
    
    begin = 0
    end = len(arr)
    while True:
        if begin == end:
            break
            
        mid = begin + end >> 1
            
        ret = compareTo(arr[mid], value)
        if ret < 0:
            begin = mid + 1
        elif ret > 0:
            end = mid
        else:
            stop = begin - 1
            begin = mid
            while True:
                pivot = -(-begin - stop >> 1)
                if pivot == begin:
                    break
                    
                if compareTo(arr[pivot], value):
                    stop = pivot
                else:
                    begin = pivot

            while True:
                pivot = mid + end >> 1
                if pivot == mid:
                    break
                    
                if compareTo(arr[pivot], value):
                    end = pivot
                else:
                    mid = pivot

            break
        
    return begin, end


def computed(prop):

    @cache
    @property
    def func(self):
        return prop(self)

    func.fget.__name__ = prop.__name__
    return func


def cache(prop):
    if isinstance(prop, property): 

        class cached(property):
                
            def __get__(self, obj, objtype=None):
                name = self.fget.__name__
                if name in obj.__dict__:
                    return obj.__dict__[name]
                value = self.fget(obj)
                obj.__dict__[name] = value
                return value

        return cached(prop.fget)
    
        
class cached_property:
    "Emulate PyProperty_Type() in Objects/descrobject.c"
 
    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc
 
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        
        name = self.fget.__name__
        if name in obj.__dict__:
            return obj.__dict__[name]
        value = self.fget(obj)
        obj.__dict__[name] = value
        return value
 
    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj, value)
 
    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(obj)
 
    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)
 
    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)
 
    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


def indexOf(arr, value):
    for i, val in enumerate(arr):
        if val == value:
            return i
    
    return -1
    
def indexFor(arr, fn):
    for i, val in enumerate(arr):
        if fn(val):
            return i
    
    return -1

class _DecoratorContextManager:
    """Allow a context manager to be used as a decorator"""

    def __call__(self, func):
        if inspect.isgeneratorfunction(func):
            return self._wrap_generator(func)

        @functools.wraps(func)
        def decorate_context(*args, **kwargs):
            with self.__class__():
                return func(*args, **kwargs)
        return decorate_context

    def _wrap_generator(self, func):
        """Wrap each generator invocation with the context manager"""
        @functools.wraps(func)
        def generator_context(*args, **kwargs):
            gen = func(*args, **kwargs)

            # Generators are suspended and unsuspended at `yield`, hence we
            # make sure the grad mode is properly set every time the execution
            # flow returns into the wrapped generator and restored when it
            # returns through our `yield` to our caller (see PR #49017).
            cls = type(self)
            try:
                # Issuing `None` to a generator fires it up
                with cls():
                    response = gen.send(None)

                while True:
                    try:
                        # Forward the response to our caller and get its next request
                        request = yield response

                    except GeneratorExit:
                        # Inform the still active generator about its imminent closure
                        with cls():
                            gen.close()
                        raise

                    except BaseException:
                        # Propagate the exception thrown at us by the caller
                        with cls():
                            response = gen.throw(*sys.exc_info())

                    else:
                        # Pass the last request to the generator and get its response
                        with cls():
                            response = gen.send(request)

            # We let the exceptions raised above by the generator's `.throw` or
            # `.send` methods bubble up to our caller, except for StopIteration
            except StopIteration as e:
                # The generator informed us that it is done: take whatever its
                # returned value (if any) was and indicate that we're done too
                # by returning it (see docs for python's return-statement).
                return e.value

        return generator_context

    
class Timer(_DecoratorContextManager):
    def __init__(self, message=None, logger=None, startHint=False):
        self.message = message
        self.logger = logger
        if startHint:
            if self.logger:
                self.logger.info(message)
            else:
                print(message)
        
    def __enter__(self):
        self.start = time.time()
        
    def __exit__(self, *_): 
        lapse = time.time() - self.start
        if self.message:
            msg = f'time cost for {self.message} is %.2f seconds' % lapse
        else:
            msg = 'time cost is %.2f seconds' % lapse
            
        if self.logger:
            self.logger.info(msg)
        else:
            print(msg)

class Timeout:
    """Timeout class using ALARM signal."""
    def __init__(self, sec):
        self.sec = sec
 
    def __enter__(self):
        signal.signal(signal.SIGALRM, self.raise_timeout)
        signal.alarm(self.sec)
 
    def __exit__(self, *args):
        signal.alarm(0)    # disable alarm

    def raise_timeout(self, *args):
        raise Exception("runtime timeout")

def setitem(arr, *args):
    *indices, index, value = args
    
    for i in indices:
        if i >= len(arr):
            arr += [None] * (i + 1 - len(arr))
            
        if arr[i] is None:
            arr[i] = []
            
        arr = arr[i]
        
    if index >= len(arr):
        arr += [None] * (index + 1 - len(arr))
        
    arr[index] = value
    
    
def getitem(data, *indices):
    for i in indices:
        if data is None:
            return
                
        if i >= len(data):
            return
        
        data = data[i]
    
    return data


def splice(arr, index, num, *args):
    assignment = []
    insertion = []

    for i, arg in enumerate(args):
        if i < num:
            assignment.append(arg)
        else:
            insertion.append(arg)

    if len(assignment) < num:
        num = len(assignment)
        arr[index: index + num] = assignment
        del arr[index + num:]
    else:
        arr[index: index + num] = assignment

        index += num
        for arg in insertion:
            arr.insert(index, arg)
            index += 1

    return arr

def split_filename(filename):
    extIndex = filename.rindex('.')
    return filename[:extIndex], filename[extIndex + 1:]


def json_to_array(json):
    if isinstance(json, list):
        return [json_to_array(data) for data in json]
    
    if isinstance(json, dict) and all(isinstance(index, int) or index.isdigit() for index in json):
        arr = []
        for index in json: 
            setitem(arr, int(index), json_to_array(json[index]))
    
        return arr
    
    return json


def toggleCase(s):
    return ''.join(ch.upper() if ch.islower() else ch.lower() if ch.isupper() else ch for ch in s)


def clip(this, min, max):
    if this < min:
        return min
    
    if this > max:
        return max
        
    return this

def deleteIndices(arr, fn, postprocess=None):
    indicesToDelete = []
    is_binary = fn.__code__.co_argcount > 1
    for i in range(len(arr)):
        if fn(arr, i) if is_binary else fn(arr[i]):
            indicesToDelete.append(i)

    if postprocess:
        is_binary = postprocess.__code__.co_argcount > 1
            
    if indicesToDelete:
        indicesToDelete.reverse()
        
        if not isinstance(arr, list):
            arr = [*arr]

        for i in indicesToDelete:
            if postprocess:
                if is_binary:
                    postprocess(arr, i)
                else: 
                    postprocess(arr[i])
            del arr[i]
            
        return arr
     
def batches(listOfElement, batch_size):
    batches = [None] * ((len(listOfElement) + batch_size - 1) // batch_size)

    index = 0
    for i in range(0, len(listOfElement), batch_size):
        batches[index] = listOfElement[i: min(i + batch_size, len(listOfElement))]
        index += 1
    return batches


def is_same(list):
    for i in range(1, len(list)):
        if list[i] != list[i - 1]:
            return False
                        
    return True


@unique
class Enum(Enum):
    def __new__(cls):
        value = len(cls.__members__)
        
        if not value:
            cls.members = []
            
        obj = object.__new__(cls)
        obj._value_ = value
        
        cls.members.append(obj)
        return obj
    
    @property
    def type(self):
        return self.name.split('_')[0]
    

def merge_sort(arr1, arr2, compareTo=None, ret=None):
    if ret == None:
        ret = Array()
    
    if compareTo == None:
        compareTo = lambda a, b: a - b
    
    _merge_sort(arr1, len(arr1), arr2, len(arr2), compareTo, ret)
    
    return ret

# precondition: the destine array is not the same as the source arrays;
def _merge_sort(arr1, sz1, arr2, sz2, compareTo, dst):
    i = 0
    j = 0
    k = 0
    while i < sz1 and j < sz2:
        if compareTo(arr1[i], arr2[j]) < 0:
            dst[k] = arr1[i]
            i += 1
        else:
            dst[k] = arr2[j]
            j += 1
            
        k += 1
    
    while i < sz1:
        dst[k] = arr1[i]
        k += 1
        i += 1
        
    while j < sz2:
        dst[k] = arr2[j]
        k += 1
        j += 1


def items(dict, **kwargs):
    [*dict] = dict.items()
    dict.sort(key=lambda args: args[0], **kwargs)
    return dict

def fromEntries(*args, object=True):
    obj = Object() if object else {}
    
    if len(args) > 1:
        for i in range(0, len(args), 2):
            obj[args[i]] = args[i + 1]
    else:
        args, = args
        for key, value in args:
            obj[key] = value

    return obj


if __name__ == '__main__':
    obj = Object(a=1, b=2, c=3)
    dic = {**obj}
    print(dic)
    
